@@ -64,10 +64,10 @@ static void he_menu_store_tree_model_init(GtkTreeModelIface *tree_model_iface);
 
 G_DEFINE_TYPE_EXTENDED(HeMenuStore, he_menu_store, G_TYPE_OBJECT, 0, G_IMPLEMENT_INTERFACE(GTK_TYPE_TREE_MODEL, he_menu_store_tree_model_init));
 
+static void menu_item_label_changed(GObject *label, GParamSpec *null, HeMenuStore *hms);
 static void menu_item_changed(GObject *menu_item, GParamSpec *null, HeMenuStore *hms);
 static void menu_item_toggled(GObject *menu_item, HeMenuStore *hms);
 
-
 /*
  * Return the flags for the model
  */
@@ -102,6 +102,7 @@ he_menu_store_get_column_type(GtkTreeModel *tm, gint index)
 	  HE_MENU_STORE_IS_CHECK_ITEM   == index ? G_TYPE_BOOLEAN  :
 	  HE_MENU_STORE_IS_RADIO_ITEM   == index ? G_TYPE_BOOLEAN  :
 		HE_MENU_STORE_TEXT            == index ? G_TYPE_STRING   :
+		HE_MENU_STORE_MARKUP          == index ? G_TYPE_STRING   :
 		HE_MENU_STORE_SENSITIVE       == index ? G_TYPE_BOOLEAN  :
 		HE_MENU_STORE_ACTIVE          == index ? G_TYPE_BOOLEAN  :
 		HE_MENU_STORE_INCONSISTENT    == index ? G_TYPE_BOOLEAN  :
@@ -288,6 +289,23 @@ image_get_pixbuf(GtkImage *image)
 	return pb;
 }
 
+static void
+set_value_from_menu_item_string(GtkWidget *menu_item, const char *(*text_retr_function)(GtkLabel *), GValue *val)
+{
+	GtkWidget *label = NULL;
+
+	_hmv_private_get_first_label_found(menu_item, &label);
+	if (label) {
+		char *text = NULL;
+
+		text = g_strdup(text_retr_function(GTK_LABEL(label)));
+		if (gtk_label_get_use_underline(GTK_LABEL(label)))
+			text = strip_underscores(text);
+		g_value_set_string(val, text);
+		g_free(text);
+	}
+}
+
 /*
  * Get a value for the given column of the given #GtkTreeIter
  */
@@ -395,15 +413,11 @@ he_menu_store_get_value(GtkTreeModel *tm, GtkTreeIter *iter, gint column, GValue
 			break;
 
 		case HE_MENU_STORE_TEXT:
-			{
-				GtkWidget *label = NULL;
-				char *text = g_strdup(_hmv_private_menu_item_get_label(GTK_WIDGET(menu_item), &label));
+			set_value_from_menu_item_string(GTK_WIDGET(menu_item), gtk_label_get_text, val);
+			break;
 
-				if (gtk_label_get_use_underline(GTK_LABEL(label)))
-					text = strip_underscores(text);
-				g_value_set_string(val, text);
-				g_free(text);
-			}
+		case HE_MENU_STORE_MARKUP:
+			set_value_from_menu_item_string(GTK_WIDGET(menu_item), gtk_label_get_label, val);
 			break;
 
 		case HE_MENU_STORE_MENU_WIDGET:
@@ -603,8 +617,13 @@ static void
 unhook_signals(GObject *menu_item, HeMenuStore *hms)
 {
 	if (GPOINTER_TO_INT(g_object_get_data(G_OBJECT(menu_item), "visited"))) {
+		GtkWidget *label = NULL;
+
+		_hmv_private_get_first_label_found(GTK_WIDGET(menu_item), &label);
 		g_object_set_data(menu_item, "visited", FALSE);
 		g_signal_handlers_disconnect_by_func(menu_item, (GCallback)menu_item_changed, hms);
+		if (label)
+			g_signal_handlers_disconnect_by_func(label, (GCallback)menu_item_label_changed, hms);
 		if (GTK_IS_RADIO_MENU_ITEM(menu_item))
 			g_signal_handlers_disconnect_by_func(menu_item, (GCallback)menu_item_toggled, hms);
 		if (hms->priv->is_deep) {
@@ -650,6 +669,26 @@ menu_item_removed(GtkContainer *container, GObject *menu_item, HeMenuStore *hms)
 	}
 }
 
+static void
+menu_item_label_changed(GObject *label, GParamSpec *null, HeMenuStore *hms)
+{
+	GObject *menu_item = NULL;
+	GtkWidget *parent = NULL;
+
+	do {
+		parent = gtk_widget_get_parent(GTK_WIDGET(label));
+		if (parent) {
+			if (GTK_IS_MENU_ITEM(parent))
+				menu_item = G_OBJECT(parent);
+		}
+		else
+			break;
+	} while (menu_item == NULL);
+
+	if (menu_item)
+		menu_item_changed(menu_item, NULL, hms);
+}
+
 /*
  * Notification of change to a given #GtkMenuItem via #GObject::notify
  */
@@ -687,6 +726,7 @@ menu_item_toggled(GObject *menu_item, HeMenuStore *hms)
 static void
 hook_up_signals(GtkMenu *menu, HeMenuStore *hms)
 {
+	GObject *label = NULL;
 	GList *ll = gtk_container_get_children(GTK_CONTAINER(menu));
 	int Nix = 0;
 
@@ -697,6 +737,10 @@ hook_up_signals(GtkMenu *menu, HeMenuStore *hms)
 					g_object_set_data(G_OBJECT(ll->data), "visited", GINT_TO_POINTER(TRUE));
 					g_object_set_data(G_OBJECT(ll->data), "idx", GINT_TO_POINTER(Nix));
 					g_signal_connect(ll->data, "notify", (GCallback)menu_item_changed, hms);
+					label = NULL;
+					_hmv_private_get_first_label_found(GTK_WIDGET(ll->data), (GtkWidget **)&label);
+					if (label)
+						g_signal_connect(label, "notify", (GCallback)menu_item_label_changed, hms);
 					if (GTK_IS_RADIO_MENU_ITEM(ll->data))
 						g_signal_connect(ll->data, "toggled", (GCallback)menu_item_toggled, hms);
 					if (hms->priv->is_deep) {
@@ -748,16 +792,16 @@ set_menu(HeMenuStore *hms, GtkMenu *menu)
 			g_object_set_data(G_OBJECT(hms->priv->menu), "he-menu-store", NULL);
 			g_signal_handlers_disconnect_by_func(G_OBJECT(hms->priv->menu), menu_item_removed, hms);
 			g_signal_handlers_disconnect_by_func(G_OBJECT(hms->priv->menu), menu_item_added, hms);
-			g_object_unref(G_OBJECT(hms->priv->menu));
 			for (; ll; ll = g_list_delete_link(ll, ll))
 				if (ll->data)
 					if (IS_VALID_MENU_ITEM(ll->data))
 						unhook_signals(G_OBJECT(ll->data), hms);
+			g_object_unref(G_OBJECT(hms->priv->menu));
+			hms->priv->menu = NULL;
 		}
 
-		hms->priv->menu = menu ? g_object_ref_sink(menu) : NULL;
-
 		if (menu) {
+			hms->priv->menu = g_object_ref_sink(menu);
 			g_object_set_data(G_OBJECT(menu), "he-menu-store", hms);
 			g_signal_connect(G_OBJECT(menu), "add", (GCallback)menu_item_added, hms);
 			g_signal_connect(G_OBJECT(menu), "remove", (GCallback)menu_item_removed, hms);
