@@ -37,6 +37,11 @@ struct _HeMenuViewColumnPrivate
 	GtkWidget *tv;
 	GtkWidget *menu_widget;
 	gboolean do_activate;
+	GObject *cr_text;
+	GtkTreeViewColumn *col_text;
+	GtkTreeModel *old_tm;
+	GtkTreeModel *old_hms;
+	int max_tv_width;
 };
 
 G_DEFINE_TYPE(HeMenuViewColumn, he_menu_view_column, HILDON_TYPE_PANNABLE_AREA);
@@ -56,17 +61,39 @@ size_request(GtkWidget *widget, GtkRequisition *rq)
 	HeMenuViewColumn *hmvc = HE_MENU_VIEW_COLUMN(widget);
 	GtkRequisition rq_tv = {.width = 0, .height = 0};
 	guint indicator_width;
+	gboolean ellipsize_set;
+
+	g_object_get(G_OBJECT(hmvc->priv->cr_text), "ellipsize-set", &ellipsize_set, NULL);
 
 	gtk_widget_style_get(widget, "indicator-width", &indicator_width, NULL);
 	gtk_widget_size_request(hmvc->priv->tv, &rq_tv);
 
+	if (rq_tv.width > hmvc->priv->max_tv_width)
+		hmvc->priv->max_tv_width = rq_tv.width;
+
+	if (ellipsize_set && rq_tv.width < hmvc->priv->max_tv_width)
+		rq_tv.width = hmvc->priv->max_tv_width;
+
 	rq->width  = 2 * GTK_CONTAINER(widget)->border_width + rq_tv.width + indicator_width;
 	rq->height = 2 * GTK_CONTAINER(widget)->border_width + rq_tv.height;
+}
 
-//	g_print("%s: size_request: width  = 2 * %3d + %3d + %3d = %3d\n", 
-//		gtk_widget_get_name(widget), GTK_CONTAINER(widget)->border_width, rq_tv.width, indicator_width, rq->width);
-//	g_print("%s: size_request: height = 2 * %3d + %3d       = %3d\n",
-//		gtk_widget_get_name(widget), GTK_CONTAINER(widget)->border_width, rq_tv.height, rq->height);
+static void
+size_allocate(GtkWidget *widget, GtkAllocation *alloc)
+{
+	HeMenuViewColumn *hmvc = HE_MENU_VIEW_COLUMN(widget);
+	gboolean ellipsize_set;
+	GtkWidgetClass *gtkwidget_class_parent = GTK_WIDGET_CLASS(he_menu_view_column_parent_class);
+
+	if (gtkwidget_class_parent->size_allocate)
+		gtkwidget_class_parent->size_allocate(widget, alloc);
+
+	g_object_get(hmvc->priv->cr_text, "ellipsize-set", &ellipsize_set, NULL);
+
+	if (widget->allocation.width < widget->requisition.width && !ellipsize_set) {
+		g_object_set(hmvc->priv->cr_text, "ellipsize", PANGO_ELLIPSIZE_END, "ellipsize-set", TRUE, NULL);
+		gtk_tree_view_columns_autosize(GTK_TREE_VIEW(hmvc->priv->tv));
+	}
 }
 
 /*
@@ -117,6 +144,12 @@ get_property(GObject *obj, guint property_id, GValue *val, GParamSpec *pspec)
 	}
 }
 
+static void
+finalize(GObject *obj)
+{
+	he_menu_view_column_set_menu(HE_MENU_VIEW_COLUMN(obj), NULL);
+}
+
 /*
  * Initialize the #HeMenuViewColumn class
  */
@@ -128,8 +161,10 @@ he_menu_view_column_class_init(HeMenuViewColumnClass *hmvc_class)
 
 	gobject_class->set_property = set_property;
 	gobject_class->get_property = get_property;
+	gobject_class->finalize     = finalize;
 
 	gtkwidget_class->size_request  = size_request;
+	gtkwidget_class->size_allocate = size_allocate;
 
 	g_object_class_install_property(gobject_class, HE_MENU_VIEW_COLUMN_MENU_PROPERTY,
 		g_param_spec_object("menu", "Menu", "The menu to display",
@@ -209,12 +244,21 @@ static void
 he_menu_view_column_init(HeMenuViewColumn *hmvc)
 {
 	GtkCellRenderer *cr;
-	GtkTreeViewColumn *col = gtk_tree_view_column_new();
+	GtkTreeViewColumn *col;
 
 	hmvc->priv = G_TYPE_INSTANCE_GET_PRIVATE(hmvc, HE_TYPE_MENU_VIEW_COLUMN, HeMenuViewColumnPrivate);
 
-	hmvc->priv->tv = g_object_new(GTK_TYPE_TREE_VIEW, "visible", TRUE, NULL);
+	hmvc->priv->old_tm = NULL;
+	hmvc->priv->old_hms = NULL;
+	hmvc->priv->menu_widget = NULL;
+	hmvc->priv->do_activate = TRUE;
+	hmvc->priv->max_tv_width = -1;
 
+	hmvc->priv->tv = g_object_new(GTK_TYPE_TREE_VIEW, "visible", TRUE, NULL);
+	g_object_add_weak_pointer(G_OBJECT(hmvc->priv->tv), (gpointer *)&(hmvc->priv->tv));
+
+	/* Pixbuf/toggle/radio column */
+	col = g_object_new(GTK_TYPE_TREE_VIEW_COLUMN, NULL);
 	cr = g_object_new(GTK_TYPE_CELL_RENDERER_PIXBUF, "xalign", 0.5, "yalign", 0.5, NULL);
 	gtk_tree_view_column_pack_start(col, cr, TRUE);
 	gtk_tree_view_column_set_attributes(col, cr,
@@ -234,28 +278,31 @@ he_menu_view_column_init(HeMenuViewColumn *hmvc)
 		NULL);
 	gtk_tree_view_append_column(GTK_TREE_VIEW(hmvc->priv->tv), col);
 
-	col = gtk_tree_view_column_new();
+	/* Label column */
+	col = g_object_new(GTK_TYPE_TREE_VIEW_COLUMN, "expand", TRUE, NULL);
 	cr = g_object_new(GTK_TYPE_CELL_RENDERER_TEXT, NULL);
 	gtk_tree_view_column_pack_start(col, cr, TRUE);
 	gtk_tree_view_column_set_attributes(col, cr,
 		"sensitive", HE_MENU_STORE_SENSITIVE,
-		"text",      HE_MENU_STORE_TEXT,
+		"markup",    HE_MENU_STORE_MARKUP,
 		NULL);
+	hmvc->priv->cr_text = G_OBJECT(cr);
+	hmvc->priv->col_text = col;
+	gtk_tree_view_append_column(GTK_TREE_VIEW(hmvc->priv->tv), col);
 
+	/* Submenu column */
+	col = g_object_new(GTK_TYPE_TREE_VIEW_COLUMN, NULL);
 	cr = g_object_new(GTK_TYPE_CELL_RENDERER_PIXBUF, "stock-id", GTK_STOCK_GO_FORWARD, NULL);
 	gtk_tree_view_column_pack_start(col, cr, FALSE);
 	gtk_tree_view_column_set_attributes(col, cr,
 		"visible",   HE_MENU_STORE_HAS_CHILDREN,
 		"sensitive", HE_MENU_STORE_SENSITIVE,
 		NULL);
-
 	gtk_tree_view_append_column(GTK_TREE_VIEW(hmvc->priv->tv), col);
+
 	gtk_container_add(GTK_CONTAINER(hmvc), hmvc->priv->tv);
 	g_signal_connect(G_OBJECT(hmvc->priv->tv), "row-activated", (GCallback)tv_row_activated, hmvc);
 	g_signal_connect(G_OBJECT(hmvc->priv->tv), "button-release-event", (GCallback)tv_button_release, hmvc);
-
-	hmvc->priv->menu_widget = NULL;
-	hmvc->priv->do_activate = TRUE;
 }
 
 /**
@@ -285,6 +332,23 @@ he_menu_view_column_get_menu(HeMenuViewColumn *hmvc)
 	return menu;
 }
 
+/*
+ * Turn off the ellipsize if it's on, because it may no longer be necessary
+ */
+static void
+reset_resize(HeMenuViewColumn *hmvc)
+{
+	gboolean ellipsize_set;
+
+	g_object_get(G_OBJECT(hmvc->priv->cr_text), "ellipsize-set", &ellipsize_set, NULL);
+
+	if (ellipsize_set) {
+		g_object_set(G_OBJECT(hmvc->priv->cr_text), "ellipsize", PANGO_ELLIPSIZE_NONE, "ellipsize-set", FALSE, NULL);
+		hmvc->priv->max_tv_width = -1;
+		gtk_tree_view_columns_autosize(GTK_TREE_VIEW(hmvc->priv->tv));
+	}
+}
+
 /**
  * he_menu_view_column_set_menu:
  * @hmvc: The #HeMenuViewColumn to display the #GtkMenu
@@ -297,21 +361,45 @@ he_menu_view_column_get_menu(HeMenuViewColumn *hmvc)
 void
 he_menu_view_column_set_menu(HeMenuViewColumn *hmvc, GtkMenu *menu)
 {
-	GObject *old_tm;
+	GtkTreeModel *tm = NULL;
+	GtkTreeModel *hms = NULL;
 
 	g_return_if_fail(HE_IS_MENU_VIEW_COLUMN(hmvc));
-	g_return_if_fail(GTK_IS_MENU(menu));
-
-	g_object_get(G_OBJECT(hmvc->priv->tv), "model", &old_tm, NULL);
+	if (menu)
+		g_return_if_fail(GTK_IS_MENU(menu));
 
 	if (menu) {
-		GtkTreeModel *hms = g_object_new(HE_TYPE_MENU_STORE, "menu", menu, NULL);
-		GtkTreeModel *tm = g_object_new(GTK_TYPE_TREE_MODEL_FILTER, "child-model", hms, NULL);
+		hms = g_object_new(HE_TYPE_MENU_STORE, "menu", menu, NULL);
+		tm = g_object_new(GTK_TYPE_TREE_MODEL_FILTER, "child-model", hms, NULL);
 
-		g_object_unref(hms);
 		gtk_tree_model_filter_set_visible_column(GTK_TREE_MODEL_FILTER(tm), HE_MENU_STORE_VISIBLE);
 		gtk_widget_set_name(GTK_WIDGET(hmvc), gtk_widget_get_name(GTK_WIDGET(menu)));
+		g_signal_connect_swapped(G_OBJECT(tm), "row-changed", (GCallback)reset_resize, hmvc);
+		g_signal_connect_swapped(G_OBJECT(tm), "row-deleted", (GCallback)reset_resize, hmvc);
+		g_signal_connect_swapped(G_OBJECT(tm), "row-inserted", (GCallback)reset_resize, hmvc);
+	}
+
+	if (hmvc->priv->tv)
 		g_object_set(G_OBJECT(hmvc->priv->tv), "model", tm, NULL);
+
+	/* This is terrible! Why doesn't the model die with the tree view?! */
+	while (hmvc->priv->old_tm)
+		g_object_unref(hmvc->priv->old_tm);
+
+	/* This is terrible! Why doesn't the model die with the tree view?! */
+	while (hmvc->priv->old_hms)
+		g_object_unref(hmvc->priv->old_hms);
+
+	hmvc->priv->old_hms = hms;
+	hmvc->priv->old_tm = tm;
+
+	if (hms) {
+		g_object_add_weak_pointer(G_OBJECT(hms), (gpointer *)&(hmvc->priv->old_hms));
+		g_object_unref(hms);
+	}
+
+	if (tm) {
+		g_object_add_weak_pointer(G_OBJECT(tm), (gpointer *)&(hmvc->priv->old_tm));
 		g_object_unref(tm);
 	}
 }
